import sys
import os
from pathlib import Path
import logging

from proai.cmdline import run
import xmltodict

log = logging.getLogger(__name__)


def adoc2xml(manuscript_dir='/home/hobs/code/tangibleai/nlpia-manuscript/manuscript'):
    cmd = f'asciidoctor -d book -b docbook5 -D {manuscript_dir}/xml {manuscript_dir}/adoc/*.adoc'
    results = run(cmd.split(' '))
    return results


def parse_xml_files(xmlpath, ext='.xml'):
    """ Read an xml file or directory of files (non-recursively) and return a nested dict

    >>> parse_xml_files(Path(DATA_DIR) / 'xml').keys()
    """
    bookdict = {}
    xmlpath = Path(xmlpath)
    if dirpath.is_dir():
        filepaths = sorted(xmlpath.listdir())
    else:
        filepaths = [dirpath]
    for f in filepaths:
        filepath = Path(dirpath) / f
        if not str(filepath).endswith(ext):
            continue
        # print(filepath)
        xmlstring = open(filepath).read()
        xmldict = {}
        try:
            xmldict = xmltodict.parse(xmlstring)
        except Exception as e:
            log.warning(f'**ParseError** -- Unable to parse "{f}":')
            log.warning(f'    {e}')
        bookdict[f] = xmldict
    return bookdict


if __name__ == "__main__":
    xmlpath = os.path.curdir
    if len(sys.argv) > 1:
        xmlpath = sys.argv[1]
    bookdict = parse_xml_files(xmlpath)
    print(bookdict.keys())
    # print(json.dumps(bookdict, indent=2))



import copy
import logging
import sys
import time

import pandas as pd
import requests
from tqdm import tqdm


log = logging.getLogger(__name__)


def load_env(filepath='.env'):
    with open(filepath) as fin:
        lines = fin.readlines()
    var_values = {}
    for line in lines:
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        try:
            name, value = line.split('=')
            name, value = name.strip().upper(), value.strip()
            value = value[1:-1] if value[0] in '"\'' and value[-1] == value[0] else value
        except ValueError:
            name, value = None, None
        if name and value:
            var_values[name.upper().strip()] = value
    return var_values


def load_manychat_answers(filepath='data/answers_613364286.csv'):
    # globals.update(read_env())
    df = pd.read_csv(filepath, parse_dates=True)
    df['finished'] = pd.to_datetime(df['finished'])
    df['timestamp'] = (df['finished'].view('int64') / 1e9).astype('int64')
    return df


def load_augmented_manychat_answers(filepath='data/answers_613364286_with_manychat_data.csv'):
    # globals.update(read_env())
    df = pd.read_csv(filepath, parse_dates=True)
    df['finished'] = pd.to_datetime(df['finished'])
    df['timestamp'] = (df['finished'].view('int64') / 1e9).astype('int64')
    return df


def find_manychat_user(row, config=None):
    if config is None:
        config = globals()
    log.info(f'Retrieving information for user named {row["full_name"]}')
    url = "https://api.manychat.com/fb/subscriber/findByName"
    params = dict(name=row['full_name'])
    headers = {
        "accept": "application/json",
        "Authorization": "Bearer Bearer {MANYCHAT_API_KEY}".format(**config)
    }
    resp = requests.get(url=url, params=params, headers=headers)
    if resp.status_code == 200:
        log.error(resp.json())
    else:
        log.error(str(resp))
    return resp


class RateLimit():
    def __init__(self, rate=13, sleep=time.sleep, sleep_seconds=1.27):
        """ rate in requests per minute """
        self.t0 = time.time()
        self.rate = rate
        self.dt = 60 / self.rate
        self.sleep = sleep
        self.sleep_seconds = sleep_seconds

    def __call__(self, sleep_seconds=None):
        self.sleep_seconds = sleep_seconds or self.sleep_seconds
        slept = 0
        while time.time() < self.t0 + self.dt:
            self.sleep(self.sleep_seconds)
            slept += 1
        self.t0 = time.time()
        return slept * self.sleep_seconds


def get_manychat_data(config=None):
    """ GET request API:
    https://api.manychat.com/swagger#/Subscriber/Manychat%5CController%5CApi%5CSubscriberController%3A%3AactionFindByName"""
    if config is None:
        config = load_env()
    df = load_manychat_answers()

    rows = []
    rate = 29  # per minute

    delta_t = 60 / rate
    t0 = time.time()
    for i, row in df.iterrows():
        row = copy.deepcopy(row.to_dict())
        row['full_name'] = copy.copy(row['user'])
        resp = find_manychat_user(row=row, config=config)
        datas = resp.json()['data']
        if len(datas) > 1:
            log.warning(f"Found {len(datas)} users with the name {row['full_name']}..")
        for data in datas:
            row = copy.deepcopy(row)
            row['user_id'] = data['id']
            for k in (
                    'status gender last_name first_name profile_pic '
                    'last_input_text phone optin_phone email optin_email subscribed '
                    'last_interaction last_seen ig_last_seen is_followup_enabled').split():
                row[k] = data[k]
            for d in data['custom_fields']:
                row[d['name']] = d['value']
            # row.update(dict(data))
            rows.append(row)

        if time.time() - t0 < delta_t:
            time.sleep(2.13)
        t0 = time.time()

    newdf = pd.DataFrame(rows)
    filepath = 'data/answers_613364286_with_manychat_data.csv'
    newdf.to_csv(filepath)
    return newdf


def upload_amplitude_record(row, config):
    url = 'https://api2.amplitude.com/2/httpapi'
    data = {
        "api_key": config['AMPLITUDE_API_KEY'],
        "events": [{
            "user_id": row['user_id'],
            "event_type": "Reported Age Group",
            "time": row['timestamp'],
            "event_properties": {
                "age_group": row['age_group']
            },
            "user_properties": {
                "age_group": row['age_group'],
            }
        }
        ]
    }
    headers = {
        'Content-Type': 'application/json',
        'Accept': '*/*'
    }
    resp = requests.post(url, json=data, params={}, headers=headers)
    if resp.status_code == 200:
        log.error(resp.json())
    else:
        log.error(str(resp))
    return resp.json()


def upload_amplitude_batch(batch, config):
    if isinstance(batch, (dict, pd.Series)):
        batch = [batch]
        log.warning(
            f'Uploading only a single event or record for {batch[0]["full_name"]} to amplitude.')
    url = 'https://api2.amplitude.com/2/httpapi'
    data = {
        "api_key": config['AMPLITUDE_API_KEY'],
        "events": [{
            "user_id": row['user_id'],
            "event_type": "Reported Age Group",
            "time": row['timestamp'],
            "event_properties": {
                "age_group": row['age_group']
            },
            "user_properties": {
                "age_group": row['age_group'],
            }
        } for row in batch]
    }
    headers = {
        'Content-Type': 'application/json',
        'Accept': '*/*'
    }
    resp = requests.post(url, json=data, params={}, headers=headers)
    print(resp)
    # TODO requests wrapper for vebosity
    if resp.status_code == 200:
        log.error(resp.json())
    else:
        log.error(str(resp))
    return resp.json()


def upload_all_amplitude_records(
        df='data/answers_613364286_with_manychat_data.csv',
        config=None,
        numevents=1,
        rate=13,
        batch_size=40,
):
    """ HTTP2 API: https://developers.amplitude.com/docs/http-api-v2#uploadrequestbody """
    if config is None:
        config = load_env()
    records = []
    if isinstance(df, str):
        df = load_augmented_manychat_answers(df)
    limit_rate = RateLimit(rate=rate)
    num_records_uploaded = 0
    total_records = len(df)
    batch = []
    numevents = min(total_records, numevents)
    for i, row in tqdm(df.iterrows(), total=int(total_records / batch_size) + 1):
        limit_rate()
        batch.append(row.to_dict())
        if len(batch) >= batch_size or num_records_uploaded == total_records - len(batch):
            print(upload_amplitude_batch(batch=batch, config=config))
            num_records_uploaded += len(batch)
            if num_records_uploaded >= total_records:
                log.error(f'All done uploading {total_records}!!!')
            batch = []
            if num_records_uploaded >= numevents:
                break
    if num_records_uploaded < total_records:
        print(f'Only uploaded {num_records_uploaded} out of {total_records} total records.')
    return pd.DataFrame(records)


if __name__ == '__main__':
    rate = 13.13
    config = load_env()
    command = config.get('DEFAULT_COMMAND', 'a')
    command_names = dict(a='amplitude-upload', m='manychat-download', b='both-manychat-and-amplitude')
    numevents = int(config.get('NUM_EVENTS', 1))
    if len(sys.argv) > 1:
        command = sys.argv[1].strip().lstrip('-').lower()[0]
    if len(sys.argv) > 2:
        numevents = int(sys.argv[2].strip().lstrip('-').lower())
    log.warning(f'Running: {command_names[command]} for {numevents} users at {rate} per min.')
    if command in 'mb':
        dfmanychat = get_manychat_data(config=config, rate=rate)
    if command in 'ab':
        dfuploaded = upload_all_amplitude_records(config=config, rate=rate, numevents=numevents)



"""Tools for persisting (pickling) machine learning pipelines to disk using `shelf`.

Alternatively, use `dill.dump_session(filepath)` and `dill.load_session(filepath)`
However, `dill.load_session` will only work if all the dependencies are importable and the same version.

Another package `joblib` is designed to work with parallelizable pipelines that work with large data.

Dependencies:
- shelve
- GitPython
"""
import shelve
from pathlib import Path
import datetime
import git

try:
    from pip._internal.operations.freeze import freeze as pip_freeze
except ImportError:  # pip < 10.0
    from pip.operations.freeze import freeze as pip_freeze

DEFAULT_EXCLUDE = ('load_globals', __name__, '__builtins__', 'sklearn', 'pd', 'pandas', 'np', 'numpy', 'shelve', 'tqdm')


def dump_session(
        filepath='python_globals_namespace.shelve',
        include=None,
        exclude=DEFAULT_EXCLUDE):
    """ Use shelf to save all variables in globals() to disk

    TODO: add a git refspec and datetime stamp to the file name and/or contents
    """
    include = include if include is not None else globals().keys()
    shelf = shelve.open(filepath, 'n')  # 'n' for new
    SHELVING_METADATA = dict(
        shelving_started=datetime.datetime.now())
    try:
        repo = git.Repo(str(Path(__file__).resolve().absolute().parent.parent))
        SHELVING_METADATA['git.Repo().active_branch'] = repo.active_branch
        SHELVING_METADATA['git.Repo().head.commit'] = repo.head.commit
        SHELVING_METADATA['git.Repo().head.abspath'] = repo.head.abspath
        SHELVING_METADATA['git.Repo().head.name'] = repo.head.name
        SHELVING_METADATA['git.Repo().head.ref'] = repo.head.ref
        SHELVING_METADATA['git.repo().head.log()[-1]'] = list(repo.head.log())[-1]
        SHELVING_METADATA['pip_freeze'] = list(pip_freeze())
    except TypeError as e:
        print(e)

    shelf['SHELVING_METADATA'] = globals()['SHELVING_METADATA']

    print(f'Shelving...\nfilepath: {filepath}\ninclude: {include}\nignore: {exclude}\n')
    for varname in include:
        if varname in exclude:
            continue
        try:
            shelf[varname] = globals()[varname]
        except (TypeError, KeyError, AttributeError) as e:
            print(f'ERROR shelving: {varname}')
            print(e)
    shelf.close()
    return shelf


def load_session(
        filepath='python_globals_namespace.shelve',
        include=None,
        exclude=DEFAULT_EXCLUDE):
    """ Use shelf to load all variables into globals() from a shelf file on disk """
    filepath = Path(filepath)
    if Path(filepath).suffix in '.dat .dir .bak'.split():
        filepath = filepath.with_suffix('')
    shelf = shelve.open(str(filepath))
    if not include:
        include = shelf.keys()
    if not exclude:
        exclude = DEFAULT_EXCLUDE
    for varname in shelf:
        if varname in exclude:
            continue
        try:
            globals()[varname] = shelf[varname]
        except (TypeError, KeyError, AttributeError) as e:
            print(f'ERROR upickling/unshelving/loading: {varname}')
            print(e)
    shelf.close()



import os
import subprocess
import logging

log = logging.getLogger(__name__)


def run(command, chdir=None):
    if chdir:
        log.info(f'Temporarily changing working directory to {chdir}')
        initial_cwd = os.getcwd()
        os.chdir(chdir)

    log.warning(f'Running: {" ".join(command)}')
    output = subprocess.run(command, capture_output=True)

    if chdir:
        os.chdir(initial_cwd)

    return {
        'stderr': output.stderr.decode("utf-8").splitlines(),
        'stdout': output.stderr.decode("utf-8").splitlines()
    }
